from __future__ import annotations
from pydantic import ValidationError
from clipstick import _tokens


class ClipStickError(Exception):
    """Base clipstick Exception"""

    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = message

    def __str__(self) -> str:
        return self.message


class InvalidModel(ClipStickError):
    """Raised when your clipstick model is invalid"""


class InvalidTypesInUnion(InvalidModel):
    def __init__(self) -> None:
        super().__init__("A union composing a subcommand must all be of type BaseModel")


class NoDefaultAllowedForSubcommand(InvalidModel):
    def __init__(self) -> None:
        super().__init__("A subcommand cannot have a default value.")


class TooManySubcommands(InvalidModel):
    def __init__(self) -> None:
        super().__init__("Only one subcommand per model allowed.")


class TooManyShortsException(ClipStickError):
    def __init__(self, model, shorts: list[str]) -> None:
        super().__init__("too many shorts defined inside model")
        self._model = model
        self._shorts = shorts

    def __str__(self):
        return f"{self.message}, model={self._model}, shorts={self._shorts}"


class FieldError(ClipStickError):
    """A pydantic validation error wrapper."""

    def __init__(
        self,
        exception: ValidationError,
        token: _tokens.Command | _tokens.Subcommand,
        provided_args: list[str],
    ) -> None:
        super().__init__("A pydantic validation wrapper")
        errors = []
        for error in exception.errors():
            input = error["input"]
            error_msg = error["msg"]
            # todo: most of times I need just the fist item.
            # Have not encountered a situation where I need something else,
            # but it will need some investigating though.
            failing_field = error["loc"][0]

            # find the token by using the input value (which is the key) that is causing the exception.
            positional_token = next(
                (tk for tk in token.args if tk.key == failing_field), None
            )
            if positional_token:
                # this token relates to a positional argument.
                if isinstance(token, _tokens.Subcommand):
                    errors.append(
                        f"Incorrect value for {positional_token.user_keys[0]} in {token.user_keys[0]}. {error_msg}, value: {input}"
                    )
                else:
                    errors.append(
                        f"Incorrect value for {positional_token.user_keys[0]}. {error_msg}, value: {input}"
                    )
            else:
                optional_token = next(
                    tk for tk in token.optional_kwargs if tk.key == failing_field
                )
                if optional_token and optional_token.indices:
                    used_token = provided_args[optional_token.indices][0]
                    # do token stuff
                    if isinstance(token, _tokens.Subcommand):
                        errors.append(
                            f"Incorrect value for {used_token} in {token.user_keys[0]}. {error_msg}, value: {input}"
                        )
                    else:
                        errors.append(
                            f"Incorrect value for {used_token}. {error_msg}, value: {input}"
                        )
        self.errors = errors

    def __str__(self) -> str:
        return "\n".join(self.errors)



from __future__ import annotations
from pydantic.fields import FieldInfo
from typing import Iterator, Literal, get_args, TYPE_CHECKING
from rich.text import Text
from rich.console import Console
from rich.table import Table

console = Console()
if TYPE_CHECKING:  # pragma: no cover
    from clipstick._tokens import Command, Subcommand


def help(command: Command | Subcommand) -> None:
    indent = 2
    min_args_width = 20
    call_stack = list(call_stack_from_tokens(command))

    entry_point = " ".join(
        ("/".join(token.user_keys) for token in reversed(call_stack))
    )

    usage_line = Text()
    usage_line.append("Usage: ", style="magenta")
    usage_line.append(entry_point)
    if command.args:
        usage_line.append(" [Arguments]")
    if command.optional_kwargs:
        usage_line.append(" [Options]")
    if command.sub_commands:
        usage_line.append(" [Subcommands]")
    console.print(usage_line)
    console.print("")
    console.print(command.cls.__doc__)

    if command.args:
        tbl = Table.grid(collapse_padding=True, padding=(0, 1))
        tbl.add_column(width=indent)  # empty column
        tbl.add_column(min_width=min_args_width)
        tbl.add_column()
        tbl.add_column()

        console.print("")
        console.print("Arguments:", style="bold")
        for arg in command.args:
            tbl.add_row(
                "",
                user_keys(arg.user_keys),
                field_description(arg.field_info),
                Text(f"[{type_from_annotation(arg.field_info)}]"),
            )
        console.print(tbl)
    if command.optional_kwargs:
        tbl = Table.grid(collapse_padding=True, padding=(0, 1))
        tbl.add_column(width=indent)  # empty column
        tbl.add_column(min_width=0)  # for the shorthands.
        tbl.add_column(min_width=min_args_width)  # keys
        tbl.add_column()  # description
        tbl.add_column()  # type
        tbl.add_column()  # default

        console.print("")
        console.print("Options:", style="bold")
        for kwarg in command.optional_kwargs:
            tbl.add_row(
                "",
                user_keys(kwarg.short_keys),
                user_keys(kwarg.keys),
                field_description(kwarg.field_info),
                Text(f"[{type_from_annotation(kwarg.field_info)}]"),
                Text(f"[default = {add_default(kwarg.field_info)}]"),
            )
        console.print(tbl)
    if command.sub_commands:
        tbl = Table.grid(collapse_padding=True, padding=(0, 1))
        tbl.add_column(width=indent)  # empty column
        tbl.add_column(min_width=min_args_width)  # commands
        tbl.add_column()  # description

        console.print("")
        console.print("Subcommands:", style="bold")

        for sub_command in command.sub_commands:
            tbl.add_row("", user_keys(sub_command.user_keys), sub_command.cls.__doc__)
        console.print(tbl)


def call_stack_from_tokens(
    token: Command | Subcommand,
) -> Iterator[Command | Subcommand]:
    """Return the sequence of subcommands the user provided to reach this specific subcommand."""
    yield token
    if token.parent is None:
        return
    yield from call_stack_from_tokens(token.parent)


def is_literal(field_info: FieldInfo) -> bool:
    return getattr(field_info.annotation, "__origin__", None) == Literal


def type_from_annotation(field_info: FieldInfo) -> str:
    if not field_info.annotation:
        return "unknown"
    if is_literal(field_info):
        options = get_args(field_info.annotation)
        return f"allowed values: {', '.join(options)}"
    return field_info.annotation.__name__


def field_description(field_info: FieldInfo) -> str:
    """Return a description for a pydantic field."""

    d = field_info.description
    if d is None:
        return ""
    return d


def add_default(field_info: FieldInfo) -> str:
    return field_info.default


def user_keys(user_keys: list[str]) -> Text:
    return Text("/".join(user_keys), style="magenta")



import ast
import inspect
from pydantic import BaseModel


def set_undefined_field_descriptions_from_var_docstrings(
    model: type[BaseModel],
) -> None:
    module = ast.parse(inspect.getsource(model))
    assert isinstance(module, ast.Module)
    class_def = module.body[0]
    assert isinstance(class_def, ast.ClassDef)
    if len(class_def.body) < 2:
        return

    for last, node in zip(class_def.body, class_def.body[1:]):
        if not (
            isinstance(last, ast.AnnAssign)
            and isinstance(last.target, ast.Name)
            and isinstance(node, ast.Expr)
        ):
            continue

        info = model.model_fields[last.target.id]
        if info.description is not None:
            continue

        doc_node = node.value
        if isinstance(doc_node, ast.Constant):
            docstring = doc_node.value  # 'regular' variable doc string
        else:
            raise NotImplementedError(doc_node)  # pragma: nocover

        info.description = docstring



from inspect import isclass
from itertools import chain
from pydantic.fields import FieldInfo
from types import UnionType
from clipstick._docstring import set_undefined_field_descriptions_from_var_docstrings
from clipstick._annotations import Short
from clipstick._tokens import (
    OptionalKeyArgs,
    PositionalArg,
    Subcommand,
    Command,
    BooleanFlag,
    OptionalBooleanFlag,
)
from clipstick._exceptions import (
    TooManyShortsException,
    InvalidTypesInUnion,
    NoDefaultAllowedForSubcommand,
    TooManySubcommands,
)

from pydantic import BaseModel


from typing import Iterator, get_args


def _is_subcommand(attribute: str, field_info: FieldInfo) -> bool:
    """Check if the field annotated as a subcommand."""
    if not isinstance(field_info.annotation, UnionType):
        return False
    args = get_args(field_info.annotation)
    if not all(issubclass(arg, BaseModel) for arg in args):
        raise InvalidTypesInUnion()
    if not field_info.is_required():
        raise NoDefaultAllowedForSubcommand()
    return True


def _is_boolean_type(field_info: FieldInfo) -> bool:
    anno = field_info.annotation
    if anno is bool:
        return True
    return False


def tokenize(model: type[BaseModel], sub_command: Subcommand | Command) -> None:
    # todo: move this somewhere else.
    set_undefined_field_descriptions_from_var_docstrings(model)
    _sub_command_found: bool = False
    for key, value in model.model_fields.items():
        if _is_subcommand(key, value):
            if _sub_command_found:
                raise TooManySubcommands()
            _sub_command_found = True
            # each result of the get_args call is a type[BaseModel]
            # which is processed as a subcommand.
            for annotated_model in get_args(value.annotation):
                new_sub_command = Subcommand(
                    key=key, cls=annotated_model, parent=sub_command
                )

                sub_command.sub_commands.append(new_sub_command)
                tokenize(annotated_model, new_sub_command)
        elif _is_boolean_type(value):
            if value.is_required():
                sub_command.args.append(BooleanFlag(key, field_info=value))
            else:
                sub_command.optional_kwargs.append(
                    OptionalBooleanFlag(key, field_info=value)
                )
        elif value.is_required():
            # becomes a positional
            sub_command.args.append(PositionalArg(key, field_info=value))
        else:
            sub_command.optional_kwargs.append(OptionalKeyArgs(key, field_info=value))


def validate_model(model: type[BaseModel]) -> None:
    """Before anything we validate the input model to see it is
    useful for cli generation.
    """
    # todo: validate only one subcommand.

    # todo: validate no model as field value.

    # check shorthands per model to be unique.
    _validate_shorts(model)


def _validate_shorts(model: type[BaseModel]) -> None:
    """Iterate over the complete cli model and validate each model of short-hand uniqueness.

    Returns:
        None if all ok

    Raises:
        ValueError when validation has failed.
    """
    for model in iter_over_model(model):
        _validate_shorts_in_model(model)


def _validate_shorts_in_model(model: type[BaseModel]):
    shorts = [
        short.short
        for short in chain(*(field.metadata for field in model.model_fields.values()))
        if isinstance(short, Short)
    ]

    unique_shorts = set(shorts)

    if len(shorts) == len(unique_shorts):
        return
    raise TooManyShortsException(model, shorts)


def iter_over_model(model: type[BaseModel]) -> Iterator[type[BaseModel]]:
    """Return all BaseModels within a provided BaseModel."""
    yield model

    for item in model.model_fields.values():
        if isclass(item.annotation) and issubclass(item.annotation, BaseModel):
            yield from iter_over_model(item.annotation)
        else:
            args = get_args(item.annotation)
            for arg in args:
                if isclass(arg) and issubclass(arg, BaseModel):
                    yield from iter_over_model(arg)



from __future__ import annotations
from dataclasses import dataclass, field
from functools import cached_property
import sys
from clipstick._annotations import Short
from typing import Generic, TypeVar
from itertools import chain
from pydantic import BaseModel, ValidationError
from pydantic.alias_generators import to_snake
from pydantic.fields import FieldInfo
from clipstick import _help
from clipstick._exceptions import FieldError

TTokenType = TypeVar("TTokenType")
TPydanticModel = TypeVar("TPydanticModel", bound=BaseModel)


def _to_false_key(key: str) -> str:
    return f"--no-{key.replace('_','-')}"


def _to_key(key: str) -> str:
    return f"--{key.replace('_','-')}"


def _to_short(key: str) -> str:
    return f"-{key}"


def _to_false_short(key: str) -> str:
    return f"-no-{key}"


@dataclass
class Token(Generic[TTokenType]):
    """Represents either a pydantic model or a pydantic field.

    A token is used to interpret provided arguments (check if there is a match) and
    if so, consume a part of provided arguments (during parsing.)
    """

    key: str

    def match(self, idx: int, values: list[str]) -> tuple[bool, int]:
        """Try to match a (range of) value(s) starting from an index.

        Matching logic is implemented depending on argument type (like positional, optional etc.)

        If a match is found it will be added to the list of tokens which are used for final parsing.
        The token stores the indices of the provided argument list to later parse the arguments.

        Returns:
            A boolean indicating the match was a success.
                For a positional (required) argument this means an explicit match.
                For an optional both an exact match will return True, but also no match will return True.
                The latter is a signal for the match to continue.
            A new index indicating the starting point for the next token match.
        """
        raise NotImplementedError()

    def parse(self, values: list[str]) -> dict[str, TTokenType]:
        """Parse data from the provided values based on the match logic implemented in this class."""
        raise NotImplementedError()

    @property
    def user_keys(self) -> list[str]:
        """Argument keys (like --verbose or --value) provided by a user to indicate a keyword or flag.

        Many times this is a list of normalized pydantic fields or provided shorthand names.
        """
        raise NotImplementedError()


@dataclass
class PositionalArg(Token[str]):
    key: str
    field_info: FieldInfo
    indices: slice | None = None

    @cached_property
    def keys(self) -> list[str]:
        return [(self.key.replace("_", "-"))]

    @property
    def user_keys(self) -> list[str]:
        return self.keys

    def match(self, idx: int, values: list[str]) -> tuple[bool, int]:
        self.indices = slice(idx, idx + 1)
        return True, idx + 1

    def parse(self, values: list[str]) -> dict[str, str]:
        if self.indices:
            return {self.key: values[self.indices][0]}
        raise ValueError("Expecting a slice object. Got None.")


@dataclass
class OptionalKeyArgs(Token[str]):
    key: str
    field_info: FieldInfo
    indices: slice | None = None

    @cached_property
    def short_keys(self) -> list[str]:
        return [
            _to_short(short.short)
            for short in self.field_info.metadata
            if isinstance(short, Short)
        ]

    @cached_property
    def keys(self) -> list[str]:
        return [_to_key(self.key)]

    @cached_property
    def user_keys(self) -> list[str]:
        return self.keys + self.short_keys

    def match(self, idx: int, values: list[str]) -> tuple[bool, int]:
        try:
            if values[idx] not in self.user_keys:
                # As this is an optional, we are returning true to continue matching.
                return True, idx
        except IndexError:
            # As this is an optional, we are returning true to continue matching.
            return True, idx

        # consume next two values
        self.indices = slice(idx, idx + 2)
        return True, idx + 2

    def parse(self, values: list[str]) -> dict[str, str]:
        if self.indices:
            return {self.key: values[self.indices][-1]}
        return {}


@dataclass
class BooleanFlag(Token[bool]):
    """A positional (required) boolean flag value."""

    key: str
    """A pydantic field key/name"""
    field_info: FieldInfo
    indices: slice | None = None

    @cached_property
    def _short_true_keys(self) -> list[str]:
        """Return a list of 'shorts' to a list of short hand arguments.

        Example:
            ['a','b'] --> ['-a','-b']
        """
        return [_to_short(short.short) for short in self.field_info.metadata]

    @cached_property
    def _short_false_keys(self) -> list[str]:
        """Return a list of 'shorts' to a list of negated short hand arguments.

        Example:
            ['a','b'] --> ['--no-a','no-b']
        """
        return [_to_false_short(short.short) for short in self.field_info.metadata]

    @cached_property
    def _true_keys(self) -> list[str]:
        """Return a list of argument keys."""
        return [_to_key(self.key)]

    @cached_property
    def _false_keys(self) -> list[str]:
        """Return a list of negated argument keys."""
        return [_to_false_key(self.key)]

    @cached_property
    def short_keys(self) -> list[str]:
        return self._short_false_keys + self._short_true_keys

    @cached_property
    def keys(self) -> list[str]:
        return self._true_keys + self._false_keys

    @property
    def user_keys(self) -> list[str]:
        return self.short_keys + self.keys

    def match(self, idx: int, values: list[str]) -> tuple[bool, int]:
        if len(values) <= idx:
            return False, idx

        if values[idx] in self.user_keys:
            self.indices = slice(idx, idx + 1)
            return True, idx + 1
        return False, idx

    def parse(self, values: list[str]) -> dict[str, bool]:
        if self.indices:
            val = values[self.indices][0] in self._true_keys + self._short_true_keys
            return {self.key: val}
        return {}


@dataclass
class OptionalBooleanFlag(BooleanFlag):
    key: str
    field_info: FieldInfo
    indices: slice | None = None

    @cached_property
    def short_keys(self) -> list[str]:
        if self.field_info.default is False:
            return self._short_true_keys
        return self._short_false_keys

    @cached_property
    def keys(self) -> list[str]:
        if self.field_info.default is False:
            return self._true_keys
        return self._false_keys

    @cached_property
    def user_keys(self) -> list[str]:
        if self.field_info.default is False:
            return self._true_keys + self._short_true_keys
        else:
            return self._false_keys + self._short_false_keys

    def match(self, idx: int, values: list[str]) -> tuple[bool, int]:
        if len(values) <= idx:
            return True, idx

        if values[idx] in self.user_keys:
            self.indices = slice(idx, idx + 1)
            return True, idx + 1
        return True, idx


@dataclass
class Command(Token[TPydanticModel]):
    """The main/base class of your CLI.

    There will be only one of this in your CLI.
    """

    key: str
    """Reference to the key in the pydantic model."""

    cls: type[TPydanticModel]
    """Pydantic class. Used for instantiating this command."""

    parent: "Command" | "Subcommand" | None
    """The full command that got you here."""

    indices: slice | None = None
    """The indices which are consumed of the provided arguments."""

    args: list[PositionalArg | BooleanFlag] = field(default_factory=list)
    """Collection of required arguments associated with this command."""

    optional_kwargs: list[OptionalKeyArgs | OptionalBooleanFlag] = field(
        default_factory=list
    )
    """Collection of optional keyword arguments associated with this command."""

    sub_commands: list["Subcommand"] = field(default_factory=list)

    @property
    def user_keys(self) -> list[str]:
        return [self.key]

    def match(self, idx: int, values: list[str]) -> tuple[bool, int]:
        """Check for token match.

        As a result the subcommand has been stripped down to a one-branch tree, meaning
        all sub_commands collections in all (nested)
        subcommands have only one or no children.


        Args:
            idx: values index to start the matching from.
            values: the list of provided arguments that need parsing

        Returns:
            tuple of bool and int.
                bool indicates whether to continue matching
                int indicates the new starting point for the next token to match.
        """
        start_idx = idx

        if _is_help_key(idx, values):
            _help.help(self)
            sys.exit(0)
        for arg in chain(self.args, self.optional_kwargs):
            success, idx = arg.match(idx, values)
            if not success:
                return False, start_idx

        if len(self.sub_commands) == 0:
            return True, idx

        subcommands: list[tuple[bool, int, Subcommand]] = []
        for sub_command in self.sub_commands:
            sub_command_start_idx = idx
            result = sub_command.match(sub_command_start_idx, values)
            subcommands.append((*result, sub_command))

        succesfull_subcommands = [
            sub_command_structure
            for sub_command_structure in subcommands
            if sub_command_structure[0] is True
        ]

        if len(succesfull_subcommands) > 1:
            raise ValueError(
                "more than one solution-tree is found. Don't know what to do now."
            )
        if len(succesfull_subcommands) == 0:
            return False, start_idx
        _, new_idx, sub_commands = succesfull_subcommands[0]
        self.sub_commands = [sub_commands]

        return True, new_idx

    def parse(self, arguments: list[str]) -> dict[str, TPydanticModel]:
        """Populate all tokens with the provided arguments."""
        args: dict[str, str | bool] = {}
        [
            args.update(parsed.parse(arguments))
            for parsed in chain(self.args, self.optional_kwargs)
        ]
        sub_commands: dict[str, BaseModel] = {}
        assert len(self.sub_commands) <= 1
        [sub_commands.update(parsed.parse(arguments)) for parsed in self.sub_commands]

        try:
            model = self.cls(**args, **sub_commands)
        except ValidationError as err:
            raise FieldError(err, token=self, provided_args=arguments)

        return {self.key: model}


@dataclass
class Subcommand(Command):
    @property
    def user_keys(self) -> list[str]:
        snaked = to_snake(self.cls.__name__)
        return [snaked.replace("_", "-")]

    def match(self, idx: int, values: list[str]) -> tuple[bool, int]:
        """Check for token match.

        As a result the subcommand has been stripped down to a one-branch tree, meaning
        all sub_commands collections in all (nested)
        subcommands have only one or no children.


        Args:
            idx: values index to start the matching from.
            values: the list of provided arguments that need parsing

        Returns:
            tuple of bool and int.
                bool indicates whether to continue matching
                int indicates the new starting point for the next token to match.
        """
        start_idx = idx
        try:
            if values[idx] not in self.user_keys:
                return False, start_idx
        except IndexError:
            return False, start_idx

        self.indices = slice(idx, idx + 1)

        idx += 1

        return super().match(idx, values)


def _is_help_key(idx, values: list[str]) -> bool:
    try:
        if values[idx] == "-h":
            return True
    except IndexError:
        return False
    return False



from dataclasses import dataclass


@dataclass(frozen=True)
class Short:
    short: str


def short(name: str) -> Short:
    """Add a short-hand name to the full argument name.

    Args:
        name: The name of the short-hand name. Provide just the letter. not the preceding dash.

    Returns:
        A Short marker inside a type annotation.
    """
    return Short(name)



from clipstick._parse import tokenize, validate_model
from clipstick._tokens import Command, TPydanticModel
from clipstick._exceptions import ClipStickError
import sys
from clipstick._help import console


def parse(model: type[TPydanticModel], args: list[str] | None = None) -> TPydanticModel:
    try:
        validate_model(model)
    except ClipStickError as err:
        console.print(err)
        sys.exit(1)
    if args is None:
        entry_point, args = sys.argv[0], sys.argv[1:]
    else:
        # Normally the first item in your sys.argv is the command/entrypoint you've entered.
        # During testing you don't provide that (only the actual arguments you enter after that).
        entry_point = "dummy-entrypoint"

    root_node = Command(key=entry_point, cls=model, parent=None)
    try:
        tokenize(model=model, sub_command=root_node)
    except ClipStickError as err:
        console.print(err)
        sys.exit(1)

    success, idx = root_node.match(0, args)
    if not idx == len(args):
        console.print("Unable to consume all provided arguments.")
        sys.exit(1)
    if not success:
        console.print("Unable to resolve the arguments into a usable command line.")
        sys.exit(1)

    try:
        parsed = root_node.parse(args)
    except ClipStickError as err:
        console.print(err)
        sys.exit(1)
    else:
        return parsed[entry_point]
